from rest_framework.test import APITestCase
from rest_framework import status
from django.urls import reverse
from django.contrib.auth.models import Group
from django.core.files.uploadedfile import SimpleUploadedFile
from django.conf import settings
from arkindex.documents.models import Corpus
from arkindex.dataimport.models import DataFile
from arkindex.users.models import User
import uuid
import os


class TestFiles(APITestCase):
    """
    Test file management
    """

    def setUp(self):
        self.corpus = Corpus.objects.create(name='Unit Tests')
        self.user = User.objects.create_user('test@test.test', 'testtest')
        self.group = Group.objects.create(name='Demo')
        self.user.groups.add(self.group)

        # Add write access
        self.user.corpus_right.create(corpus=self.corpus, can_write=True)

    def test_file_list(self):
        df = DataFile.objects.create(
            name='test.txt', size=42, hash='aaaa', content_type='text/plain', corpus=self.corpus)
        self.client.force_login(self.user)

        response = self.client.get(reverse('api:file-list', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIn('results', data)
        self.assertEqual(len(data['results']), 1)

        file = data['results'][0]
        self.assertIn('id', file)
        self.assertIn('name', file)
        self.assertIn('size', file)
        self.assertEqual(file['id'], str(df.id))
        self.assertEqual(file['name'], df.name)
        self.assertEqual(file['size'], df.size)

    def test_file_list_requires_login(self):
        response = self.client.get(reverse('api:file-list', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_file_upload_ro(self):
        '''
        An upload on a read only corpus should fail
        '''
        public = Corpus.objects.create(name='Unit Tests', public=True)
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        self.client.force_login(self.user)

        response = self.client.post(reverse('api:file-upload', kwargs={'pk': public.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_file_upload(self):
        """
        Assert a file upload creates a database instance and saves the file
        """
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        self.client.force_login(self.user)

        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)

        df = DataFile.objects.get(id=data['id'])
        self.assertEqual(df.name, 'test.txt')
        self.assertEqual(df.size, 19)
        self.assertEqual(df.content_type, 'text/plain')
        self.assertEqual(df.corpus, self.corpus)
        self.assertTrue(os.path.exists(os.path.join(settings.MEDIA_ROOT, str(df.id))))

    def test_file_upload_requires_login(self):
        """
        Assert a file upload requires a login
        """
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_file_upload_unknown_corpus(self):
        """
        Assert file upload fails on an unknown corpus with HTTP 400
        """
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': str(uuid.uuid4())}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_file_upload_wrong_content_type(self):
        """
        Assert file upload does not trust the client defined content type
        """
        f = SimpleUploadedFile('test.mp3', b'This actually is a text file', content_type='audio/mpeg')
        self.client.force_login(self.user)

        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)

        df = DataFile.objects.get(id=data['id'])
        self.assertEqual(df.name, 'test.mp3')
        self.assertEqual(df.content_type, 'text/plain')

    def test_file_upload_unique(self):
        """
        Assert uploading the same file twice fails
        """
        self.client.force_login(self.user)

        f = SimpleUploadedFile('test.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        f = SimpleUploadedFile('test2.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_file_upload_demo_limit(self):
        """
        Assert a file upload fails for demo users when its size exceeds the corpus limit
        """
        # Create a file one byte under the allowed size
        DataFile.objects.create(
            name='test.txt', size=52428799, hash='aaaa', content_type='text/plain', corpus=self.corpus)
        # Another file of morethan one byte in size
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
