from rest_framework import status
from rest_framework.serializers import JSONField
from django.urls import reverse
from arkindex.dataimport.models import \
    DataImport, DataImportMode, DataImportState, DataImportFailure, DataFile, Revision, Repository
from arkindex.dataimport.serializers import DataImportSerializer, ImagesPayloadSerializer
from arkindex.documents.models import Corpus, Element, ElementType
from arkindex.project.tests import RedisMockAPITestCase
from arkindex.users.models import User, OAuthCredentials, CorpusRight
import uuid


class TestImports(RedisMockAPITestCase):
    """
    Test data imports management
    """

    def setUp(self):
        super().setUp()
        self.corpus = Corpus.objects.create(name='Unit Tests')
        self.user = User.objects.create_user('test@test.test', 'testtest')
        CorpusRight.objects.create(corpus=self.corpus, user=self.user, can_write=True, can_admin=True)
        self.dataimport = DataImport.objects.create(
            creator=self.user, corpus=self.corpus, mode=DataImportMode.Images)
        self.demo_volume = Element.objects.create(type=ElementType.Volume, name='Demo', corpus=self.corpus)

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        result = data['results'][0]
        self.assertEqual(result['id'], str(self.dataimport.id))
        self.assertEqual(result['state'], DataImportState.Created.value)
        self.assertEqual(result['mode'], DataImportMode.Images.value)
        self.assertEqual(result['corpus'], str(self.corpus.id))

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.dataimport.id))

    def test_create_requires_login(self):
        response = self.client.post(reverse('api:import-list'), data={
            'corpus': self.corpus.id,
            'mode': DataImportMode.Images.value
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-list'), data={
            'corpus': self.corpus.id,
            'mode': DataImportMode.Images.value
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()

        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.creator, self.user)
        self.assertEqual(dataimport.corpus, self.corpus)
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertEqual(dataimport.state, DataImportState.Created)

    def test_payload_serializer(self):
        """
        Test the DataImportSerializer will use ImagesPayloadSerializer instead of JSONField for Images workflows
        """
        self.assertIsInstance(DataImportSerializer().fields['payload'], JSONField)
        self.assertIsInstance(
            DataImportSerializer(instance=self.dataimport).fields['payload'],
            ImagesPayloadSerializer,
        )

    def test_configure_requires_login(self):
        response = self.client.put(reverse('api:import-details', kwargs={'pk': self.dataimport.id}), data={
            'corpus': str(self.dataimport.corpus.id),
            'mode': DataImportMode.Images.value,
            'state': DataImportState.Created.value,
            'payload': {
                'folder_name': 'dir',
                'volume_id': str(uuid.uuid4()),
            },
            'files': []
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_configure_no_files(self):
        """
        Test configuring an already created images workflow with no files
        """
        self.client.force_login(self.user)
        response = self.client.put(reverse('api:import-details', kwargs={'pk': self.dataimport.id}), data={
            'corpus': str(self.dataimport.corpus.id),
            'mode': DataImportMode.Images.value,
            'state': DataImportState.Created.value,
            'payload': {
                'folder_name': 'dir',
                'volume_id': str(uuid.uuid4()),
            },
            'files': [],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_configure(self):
        self.client.force_login(self.user)
        df = DataFile.objects.create(
            name='test.txt', size=42, hash='aaaa', content_type='text/plain', corpus=self.corpus)

        response = self.client.put(reverse('api:import-details', kwargs={'pk': self.dataimport.id}), data={
            'corpus': str(self.dataimport.corpus.id),
            'mode': DataImportMode.Images.value,
            'state': DataImportState.Created.value,
            'payload': {
                'folder_name': 'dir',
                'volume_id': str(self.demo_volume.id)
            },
            'files': [df.id, ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, DataImportState.Configured)
        self.assertEqual(self.dataimport.payload['folder_name'], 'dir')
        self.assertEqual(self.dataimport.folder_name, 'dir')
        self.assertTrue(self.dataimport.iiif_path.endswith('dir'))
        self.assertEqual(self.dataimport.payload['volume_id'], str(self.demo_volume.id))
        self.assertListEqual(list(self.dataimport.files.all()), [df])

    def test_start(self):
        self.client.force_login(self.user)
        df = DataFile.objects.create(
            name='test.txt', size=42, hash='aaaa', content_type='text/plain', corpus=self.corpus)
        self.dataimport.payload = {'folder_name': 'dir', 'volume_id': 'vol'}
        self.dataimport.files.add(df)
        self.dataimport.state = DataImportState.Configured
        self.dataimport.save()

        self.assertGreaterEqual(self.redis.llen('celery'), 0)

        response = self.client.put(reverse('api:import-details', kwargs={'pk': self.dataimport.id}), data={
            'corpus': str(self.dataimport.corpus.id),
            'mode': DataImportMode.Images.value,
            'state': DataImportState.Running.value,
            'payload': {
                'folder_name': 'dir',
                'volume_id': 'vol',
            },
            'files': [df.id, ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertIsNotNone(self.dataimport.root_id)

        self.assertGreaterEqual(self.redis.llen('celery'), 1)

    def test_start_demo_requires_login(self):
        df = DataFile.objects.create(
            name='test.txt', size=42, hash='aaaa', content_type='text/plain', corpus=self.corpus)
        response = self.client.post(reverse('api:import-demo', kwargs={'pk': df.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_start_demo(self):
        self.client.force_login(self.user)
        df = DataFile.objects.create(
            name='test.txt', size=42, hash='aaaa', content_type='text/plain', corpus=self.corpus)
        response = self.client.post(reverse('api:import-demo', kwargs={'pk': df.id}))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertEqual(dataimport.state, DataImportState.Running)
        self.assertListEqual(list(dataimport.files.all()), [df])
        self.assertEqual(dataimport.payload['volume_id'], str(self.demo_volume.id))
        self.assertEqual(dataimport.payload['folder_name'], 'user-{}-demo'.format(self.user.id))
        self.assertGreaterEqual(self.redis.llen('celery'), 1)

    def test_failures_require_login(self):
        self.dataimport.failures.create(path='path/to/file', message='something')
        response = self.client.get(reverse('api:import-failures', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_failures(self):
        self.client.force_login(self.user)
        failure = self.dataimport.failures.create(path='path/to/file', message='something')
        response = self.client.get(reverse('api:import-failures', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(failure.id))

    def test_save_failures(self):
        """
        Check DataImport properly saves failures
        """
        rev = Revision.objects.create(
            repo=Repository.objects.create(
                url='http://repo',
                hook_token='token',
                corpus=self.corpus,
                credentials=OAuthCredentials.objects.create(
                    user=self.user,
                    provider_name='provider',
                    provider_url='https://somewhere',
                ),
            ),
            hash='42',
            ref='ref/heads/master',
            message='a',
            author='me',
        )
        self.dataimport.revision = rev
        self.dataimport.save_failures([
            DataImportFailure(
                path='somewhere/something',
                message="Test failure",
            )
        ])
        self.assertEqual(self.dataimport.failures.count(), 1)
        self.assertEqual(self.dataimport.failures.get().message, "Test failure")
