from rest_framework.generics import ListAPIView, RetrieveAPIView
from django.db.models import prefetch_related_objects
from arkindex.documents.serializers.light import PageLightSerializer
from arkindex.documents.serializers.elements import \
    ElementSerializer, CorpusSerializer, PageSerializer, ActSerializer, SurfaceSerializer
from arkindex.documents.models import Element, ElementType, Page, Act, Corpus
from arkindex.project.mixins import CorpusACLMixin


class ElementsList(CorpusACLMixin, ListAPIView):
    """
    List documents using pagination
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        filters = {
            'type': self.request.query_params.get('type', ElementType.Volume),
        }
        if 'corpus' in self.request.query_params:
            filters['corpus'] = self.get_corpus(self.request.query_params['corpus'])
        else:
            filters['corpus__in'] = Corpus.objects.readable(self.request.user)
        return Element.objects.filter(**filters).prefetch_related('corpus').order_by('corpus', 'type', 'name')


class ElementRetrieve(RetrieveAPIView):
    """
    Retrieve a single element
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        return Element.objects.filter(corpus__in=Corpus.objects.readable(self.request.user)).select_related('corpus')


class RelatedElementsList(ListAPIView):
    """
    List all documents linked to a specified document
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        filtering = {
            'corpus__in': Corpus.objects.readable(self.request.user),
        }
        type_param = self.request.query_params.get('type')
        if type_param is not None:
            filtering['type'] = type_param
        related = Element.objects.get_related(self.kwargs['pk'], **filtering)
        prefetch_related_objects(related, 'corpus')
        return related


class ElementPages(ListAPIView):
    """
    List pages inside an element
    """
    serializer_class = PageLightSerializer

    def get_queryset(self):
        return Page.objects.get_descending(
            self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user),
            zone__isnull=False,
        ).prefetch_related('zone__image__server')


class ElementSurfaces(ListAPIView):
    """
    List surfaces inside a volume
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        return Element.objects.get_descending(
            self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user),
        ).filter(type=ElementType.Surface)


class CorpusList(ListAPIView):
    """
    List all available corpora
    """
    serializer_class = CorpusSerializer
    pagination_class = None

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)


class CorpusPages(CorpusACLMixin, ListAPIView):
    """
    List all pages in a corpus
    """
    serializer_class = PageLightSerializer

    def get_queryset(self):
        return Page.objects.filter(corpus=self.get_corpus(self.kwargs['pk'])).select_related('zone__image__server')


class PageDetails(RetrieveAPIView):
    """
    Get details for a specific page
    """
    serializer_class = PageSerializer

    def get_queryset(self):
        return Page.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user)
        ).select_related('zone__image__server')

    def get_object(self):
        page = super().get_object()
        page.parent_elements = Element.objects.get_ascending_paths(page.id)
        return page


class SurfaceDetails(RetrieveAPIView):
    """
    Get details for a specific surface
    """
    serializer_class = SurfaceSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=ElementType.Surface,
            corpus__in=Corpus.objects.readable(self.request.user),
        ).select_related('zone__image__server')

    def get_object(self):
        surf = super().get_object()
        surf.page = Page.objects.filter(zone__image=surf.zone.image) \
                                .select_related('zone__image__server').first()
        return surf


# TODO: Edition disabled
class ActEdit(RetrieveAPIView):
    """
    Edit acts and associated ElementPaths
    """
    serializer_class = ActSerializer

    def get_queryset(self):
        return Act.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def get_object(self):
        act = super().get_object()
        act.parent_elements = Element.objects.get_ascending_paths(act.id)
        act.child_elements = Element.objects.get_descending(act.id)
        return act
