from django.conf import settings
from rest_framework import status
from rest_framework.generics import CreateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from arkindex.documents.models import Transcription, TranscriptionType
from arkindex.documents.serializers.transcriptions import TranscriptionsSerializer, TranscriptionCreateSerializer
from arkindex.documents.indexer import Indexer
from arkindex.images.models import Zone
from arkindex.images.importer import bulk_transcriptions
from arkindex.project.polygon import Polygon


class TranscriptionBulk(CreateAPIView):
    '''
    Create transcriptions in bulk, all linked to the same image
    and parent element
    '''
    serializer_class = TranscriptionsSerializer
    permission_classes = (IsAuthenticated, )

    def perform_create(self, serializer):

        parent = serializer.validated_data['parent']
        image = serializer.validated_data['image']
        transcriptions = bulk_transcriptions(
            image=image,
            parent=parent,
            items=[
                {
                    'polygon': tr['polygon'],
                    'text': tr['text'],
                    'score': tr['score'],
                    'type': tr['type'],
                }
                for tr in serializer.validated_data['transcriptions']
            ],
        )

        # Reindex all created
        if transcriptions:
            Indexer().run_index(
                settings.ES_INDEX_TRANSCRIPTIONS,
                Transcription.INDEX_TYPE,
                Transcription.objects.filter(
                    id__in=[t.id for t in transcriptions]
                ),
            )


class TranscriptionCreate(CreateAPIView):
    serializer_class = TranscriptionCreateSerializer
    permission_classes = (IsAuthenticated, )

    def perform_create(self, serializer):
        transcription_type = serializer.validated_data['type']
        if transcription_type not in TranscriptionType:
            raise ValidationError("This endpoint can only import transcriptions.")
        element = serializer.validated_data['element']

        # Create a zone on the page's image
        polygon = Polygon(serializer.validated_data['polygon'])
        try:
            ts_zone, _ = Zone.objects.get_or_create(
                image=element.zone.image,
                polygon=polygon,
            )
        except Zone.MultipleObjectsReturned:
            # This should not happend, but we cannot set a unique
            # on polygon PG fields
            ts_zone = Zone.objects.filter(
                image=element.zone.image,
                polygon=polygon,
            ).first()

        ts, created = Transcription.objects.get_or_create(
            element=element,
            type=transcription_type,
            zone=ts_zone,
            text=serializer.validated_data['text'],
            defaults={
                'score': serializer.validated_data['score']
            }
        )
        if not created or ts.score != serializer.validated_data['score']:
            ts.score = serializer.validated_data['score']
            ts.save()

        # Index in ES
        Indexer().run_index(
            settings.ES_INDEX_TRANSCRIPTIONS,
            Transcription.INDEX_TYPE,
            Transcription.objects.filter(id=ts.id)
        )
        return ts

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response({'id': obj.id}, status=status.HTTP_201_CREATED, headers=headers)
