#!/usr/bin/env python3
"""
Import Himanis volumes from CSV file
"""
from django.core.management.base import BaseCommand
from arkindex.documents.tasks import import_manifests_annotations
import logging
import csv

logging.basicConfig(
    format='[%(levelname)s] %(processName)-9s: %(message)s',
    level=logging.INFO
)
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Import Himanis volumes from CSV file as an asynchronous task."

    def add_arguments(self, parser):
        parser.add_argument(
            'csv',
            help='CSV path with manifests + annotations',
            type=open,
        )
        parser.add_argument(
            '-i', '--index-root',
            help='Root folder for indexes (/home/data/indexes)',
            default='.',
        )
        parser.add_argument(
            '--corpus-id',
            required=True,
            help='Slug of corpus to import volumes into'
        )
        parser.add_argument(
            '--col-name',
            help='Index of the volume name column',
            type=int,
            default=0,
        )
        parser.add_argument(
            '--col-url',
            help='Index of the manifest URL column',
            type=int,
            default=1,
        )
        parser.add_argument(
            '--col-path',
            help='Index of the index files path column',
            type=int,
            default=2,
        )
        parser.add_argument(
            '--col-source',
            help='Index of the index source column',
            type=int,
            default=3,
        )

    def handle(self, *args, **options):
        # Handle verbosity level
        verbosity = int(options['verbosity'])
        if verbosity > 1:
            logger.setLevel(logging.DEBUG)

        # Read CSV from management file
        logger.debug('Opening CSV: {}'.format(options['csv'].name))
        csvreader = csv.reader(options['csv'])
        next(csvreader)  # skip header
        csv_data = list(filter(None, csvreader))
        assert len(csv_data) > 0, 'No data in csv !'

        task = import_manifests_annotations.delay(
            csv_data,
            options['index_root'],
            options['corpus_id'],
            col_name=options['col_name'],
            col_url=options['col_url'],
            col_path=options['col_path'],
            col_source=options['col_source'],
        )
        logger.info("Task created with UUID {}".format(task.id))
