#!/usr/bin/env python3
from django.core.management.base import BaseCommand, CommandError
from arkindex.images.importer import IndexImporter
from arkindex.documents.models import Element, ElementType
from arkindex.documents.tasks import import_annotations
import logging
import re

logging.basicConfig(
    level=logging.INFO,
    format='[%(levelname)s] %(message)s',
)


class Command(BaseCommand):
    help = 'Import annotations from index files (.idx.gz) as async task'

    def add_arguments(self, parser):
        parser.add_argument(
            'index_folder',
            help='Folder to recursively search for index files',
            default='.',
        )
        parser.add_argument(
            '--volume',
            help='ID or exact name of the volume to import annotations in',
            required=True,
        )
        parser.add_argument(
            '--dry-run',
            help='Only show the associated images found for all index files, without importing anything'
                 'This will run directly in the console and not as a task',
            action='store_true',
            default=False,
        )
        parser.add_argument(
            '--mask',
            help="""A mask to identify images from the index file path. Cannot be used with --regex.
                 Syntax: "something<PATH>something"
                 <PATH> will be used as the image path.""",
        )
        parser.add_argument(
            '--regex',
            help="""A regex to use as a mask for more complex cases.
                 Must have only one capturing group. Cannot be used with --mask.""",
            default=IndexImporter.DEFAULT_MASK,
        )

    def handle(self, *args, **options):
        # Handle verbosity level
        verbosity = int(options['verbosity'])
        root_logger = logging.getLogger('')
        if verbosity > 1:
            root_logger.setLevel(logging.DEBUG)

        # Find volume
        try:
            volume = Element.objects.get(pk=options['volume'], type=ElementType.Volume)
        except Exception:
            volume = Element.objects.get(name=options['volume'], type=ElementType.Volume)

        # Handle mask
        if options['mask'] is not None and options['regex'] != IndexImporter.DEFAULT_MASK:
            raise CommandError('--mask and --regex cannot be used simultaneously.')

        mask_regex = options['regex']

        if options['mask'] is not None:
            mask = options['mask']
            assert mask.count('<PATH>') == 1
            # Replace <PATH> with (.+) and escape the rest
            mask_regex = '^' + r'(.+)'.join(re.escape(p) for p in mask.split('<PATH>')) + '$'

        if options['dry_run']:
            importer = IndexImporter(options['index_folder'], volume, mask=mask_regex)
            importer.dry_run()
        else:
            task = import_annotations.delay(options['index_folder'], volume.id, mask=mask_regex)
            print("Task started with ID {}".format(task.id))
