#!/usr/bin/env python3
from django.core.management.base import BaseCommand
from arkindex.documents.models import Element
from arkindex.documents.surface import parse_himanis_volume_name
from arkindex.documents.surface_link import SurfaceLinker
from arkindex.documents.tasks import link_acts_surfaces
from celery import chain
import logging
import argparse
import csv

logging.basicConfig(
    format='[%(levelname)s] %(message)s',
    level=logging.INFO
)


class Command(BaseCommand):
    help = "Create acts and link their surfaces from a CSV file"

    def add_arguments(self, parser):
        parser.add_argument(
            'csv',
            help='Path to CSV file containing surface-act associations.',
            type=argparse.FileType('r'),
        )
        parser.add_argument(
            '--dry-run',
            action='store_true',
            default=False,
            help='Print matched images for each surface file without importing'
        )
        parser.add_argument(
            '--col-volume',
            default=0,
            type=int,
            help='Column index for volume'
        )
        parser.add_argument(
            '--col-number',
            default=1,
            type=int,
            help='Column index for number'
        )
        parser.add_argument(
            '--col-folio',
            default=2,
            type=int,
            help='Column index for folio'
        )

    def handle(self, *args, **options):
        # Handle verbosity level
        verbosity = int(options['verbosity'])
        root_logger = logging.getLogger('')
        if verbosity > 1:
            root_logger.setLevel(logging.DEBUG)
        logger = logging.getLogger(__name__)

        logger.debug('Loading CSV file')
        reader = csv.reader(options['csv'], delimiter=';')
        next(reader)  # Skip header line
        rows = [
            (
                row[options['col_volume']],
                row[options['col_number']],
                row[options['col_folio']]
            )
            for row in reader
            if row and len(row[0]) > 0
        ]

        logger.info('Finding volumes')
        volume_names = set(row[0] for row in rows)
        volumes = {}
        for volume_name in volume_names:
            try:
                volumes[volume_name] = parse_himanis_volume_name(volume_name)
                logger.debug("Matched volume '{}' for volume name '{}'".format(volumes[volume_name], volume_name))
            except ValueError as e:
                logger.warning('{}\t{}'.format(volume_name, str(e)))
            except Element.DoesNotExist:
                logger.warning('{}\tNo matching volume found'.format(volume_name))
            except Element.MultipleObjectsReturned:
                logger.warning('{}\tMultiple matching volumes found'.format(volume_name))
        if len(volumes) < 1:
            raise ValueError("No matched volumes.")

        logger.info('Running linkers')
        tasks = []
        for volume_name, volume in volumes.items():
            lines = list(map(lambda x: x[1:3], filter(lambda row: row[0] == volume_name, rows)))
            if options['dry_run']:
                SurfaceLinker(volume=volume, data=lines).dry_run()
            else:
                tasks.append(
                    link_acts_surfaces.si(volume_id=volume.id, data=lines)
                )

        # Start async tasks
        if tasks:
            task = chain(*tasks).delay()
            logger.info('Async task started with ID {}'.format(task.id))
