from abc import ABC, abstractmethod
from django.conf import settings
from rest_framework import serializers
from arkindex.documents.models import Element, ElementType, Act, Page, Transcription
from arkindex.project.tools import build_absolute_url


class AnnotationSerializer(ABC, serializers.BaseSerializer):
    """
    Serialize an element into a IIIF annotation
    """

    @abstractmethod
    def get_url(self, element):
        """Get an element's ID URL"""

    def get_resource(self, element):
        """Get the annotation resource for a given element."""
        return {
            "@id": self.get_url(),
            "@type": "cnt:ContentAsText",
            "chars": element.name,
            "format": "text/plain"
        }

    def get_target(self, element):
        """Get the target canvas (`on` property) for a given element."""
        assert hasattr(element, 'zone')
        return "{0}#xywh={1.x},{1.y},{1.width},{1.height}".format(
            element.zone.image.get_thumbnail_url(max_width=None, max_height=None),
            element.zone.polygon)

    def to_representation(self, element):
        assert isinstance(element, (Element, Transcription))
        assert 'request' in self.context, "A request is required to generate absolute URLs"
        return {
            "@id": self.get_url(element),
            "@type": "oa:Annotation",
            "motivation": "sc:painting",
            "on": self.get_target(element),
            "resource": self.get_resource(element),
        }


class TranscriptionAnnotationSerializer(AnnotationSerializer):
    """
    Serialize a transcription into a IIIF annotation
    """

    def get_url(self, ts):
        return build_absolute_url(
            ts,
            self.context['request'],
            'api:transcription-manifest',
            id_argument='page_pk',
            transcription_pk=ts.id,
        )

    def get_resource(self, ts):
        return {
            "@id": self.get_url(ts),
            "@type": "cnt:ContentAsText",
            "chars": ts.text,
            "format": "text/plain"
        }


class TranscriptionSearchAnnotationSerializer(TranscriptionAnnotationSerializer):

    def get_target(self, ts):
        assert isinstance(ts, Transcription)
        url = build_absolute_url(ts.element, self.context['request'], 'api:canvas-manifest')
        return "{0}#xywh={1.x},{1.y},{1.width},{1.height}".format(url, ts.zone.polygon)


class SurfaceAnnotationSerializer(AnnotationSerializer):
    """
    Serialize a surface into a IIIF annotation
    """

    def get_url(self, surface):
        return build_absolute_url(
            surface,
            self.context['request'],
            'api:surface-manifest',
            id_argument='page_pk',
            surface_pk=surface.id,
        )

    def get_resource(self, surface):
        act = Act.objects.get_ascending(surface.id)[0]
        chars = 'Act {}'.format(act.number) if act is not None else 'Unknown act'
        return {
            "@id": self.get_url(surface),
            "@type": "cnt:ContentAsText",
            "chars": chars,
            "format": "text/plain"
        }


class AnnotationListSerializer(ABC, serializers.BaseSerializer):
    """
    Serialize a list of serialized annotations into a IIIF annotation list
    """

    annotation_serializer = AnnotationSerializer

    def to_representation(self, obj):
        assert 'request' in self.context, "A request is required to generate absolute URLs"

        return {
            "@context": settings.IIIF_PRESENTATION_CONTEXT,
            "@id": self.context['request'].build_absolute_uri(),
            "@type": "sc:AnnotationList",
            "resources": self.annotation_serializer(
                self.get_elements(obj),
                context=self.context,
                many=True
            ).data
        }

    @abstractmethod
    def get_elements(self, obj):
        "Get a list of elements to serialize as annotations."


class PageAnnotationListSerializer(AnnotationListSerializer):
    """
    Serialize a page's transcriptions into a IIIF annotation list
    """

    annotation_serializer = TranscriptionAnnotationSerializer

    def get_elements(self, page):
        assert isinstance(page, Page)
        return page.transcriptions.all()


class PageActAnnotationListSerializer(AnnotationListSerializer):
    """
    Serialize an page's acts into a IIIF annotation list
    """

    annotation_serializer = SurfaceAnnotationSerializer

    def get_elements(self, page):
        assert isinstance(page, Page)
        if page.zone is None:
            return []
        return Element.objects.filter(type=ElementType.Surface, zone__image=page.zone.image)


class SurfaceAnnotationListSerializer(AnnotationListSerializer):
    """
    Serialize a single surface into a IIIF annotation list
    """

    annotation_serializer = SurfaceAnnotationSerializer

    def get_elements(self, surface):
        assert isinstance(surface, Element) and surface.type == ElementType.Surface
        return [surface]

    def to_representation(self, surface):
        serialized = super().to_representation(surface)
        query = self.context['request'].query_params.get('q')
        # When there's a search query, add all transcriptions that contain the query
        # as serialized annotations in the list's resources
        if query:
            serialized['resources'].extend(
                TranscriptionAnnotationSerializer(
                    Transcription.objects.filter(
                        zone__image=surface.zone.image,
                        zone__polygon__in=surface.zone.polygon,
                        text__icontains=query,
                        score__gte=0.5,
                    ),
                    context=self.context,
                    many=True
                ).data
            )
        return serialized


class SearchAnnotationListSerializer(AnnotationListSerializer):
    """
    Serialize a list of serialized annotations into a search result annotation list
    """

    def to_representation(self, obj):
        serialized = super().to_representation(obj)
        serialized['@context'] = settings.IIIF_SEARCH_CONTEXT
        serialized['within'] = {
            "@type": "sc:Layer",
            "total": len(serialized['resources']),
        }
        serialized['startIndex'] = 0
        serialized['hits'] = [
            {
                "@type": "search:Hit",
                "annotations": [anno['@id']],
                "match": self.get_match(anno),
            }
            for anno in serialized['resources']
        ]
        return serialized

    def get_match(self, anno):
        """Get a match text for an annotation.
        This is optional in the Search API but mandatory with Mirador"""
        return anno['resource']['chars']


class TranscriptionSearchAnnotationListSerializer(SearchAnnotationListSerializer):
    """
    Serialize a transcription search result into an annotation list
    """

    annotation_serializer = TranscriptionSearchAnnotationSerializer

    def get_elements(self, obj):
        return obj
