from arkindex.documents.models import ElementType, Element, Page, PageDirection, PageType
from arkindex.images.models import Zone
from django.db.models import Q
from django.core.exceptions import MultipleObjectsReturned
from collections import namedtuple
import xml.etree.ElementTree as ElementTree
import re
import os
import fnmatch
import csv
import logging


logger = logging.getLogger(__name__)

NS_PAGE = 'http://schema.primaresearch.org/PAGE/gts/pagecontent/2013-07-15'
NS = {'page': NS_PAGE}
NS_DELETE = "{%s}" % NS_PAGE

REGEX_VOLUME_NAME = re.compile(r'(?:JJ)?([0-9]+)([A-Z]?)')
REGEX_ACT_NUMBER = re.compile(r'([0-9]+)\s*([a-z]*)')
REGEX_FOLIO = re.compile(r'(?:fol\.?\s*)?([0-9]+)\s*(bis)?\s*(r|v)?', re.IGNORECASE)
REGEX_FOLIO_GROUP = re.compile(r'([^-_à]+)', re.IGNORECASE)


class SurfaceParser(object):
    def __init__(self, path):

        self.path = path
        self.tree = ElementTree.parse(path)

        # Check we have a page
        self.page = self.tree.find('page:Page', NS)
        assert self.page is not None, 'Missing page'

    def get_metadata(self):

        # From metadata
        metadata = self.tree.find('page:Metadata', NS)
        assert metadata is not None, 'Missing metadata'
        dictMetadata = {node.tag.replace(NS_DELETE, "").lower(): node.text for node in metadata}

        # From page
        dictMetadata.update(self.page.attrib)

        return dictMetadata

    def list_surfaces(self):
        region_indexes = None

        # Load base elements from page
        read = self.page.find('page:ReadingOrder', NS)
        if read is not None:
            order = read.find('page:OrderedGroup', NS)
            assert order is not None, 'Missing ordered group'

            # Find indexes for each region
            region_indexes = {
                refindex.get('regionRef'): int(refindex.get('index'))
                for refindex in order.findall('page:RegionRefIndexed', NS)
            }

        out = []
        for i, textRegion in enumerate(self.page.findall('page:TextRegion', NS)):

            # Build coordinates polygon
            coords = textRegion.find('page:Coords', NS)
            assert coords is not None, 'Missing coords'
            polygon = [
                (int(x), int(y))
                for x, y in re.findall('(\d+),(\d+)', coords.get('points'))
            ]

            text_id = textRegion.get('id')
            if region_indexes is not None:  # If ReadingOrder exists
                # Find index for this text region
                index = region_indexes.get(text_id)
                assert index is not None, 'Missing index in region_indexes'
            else:
                index = i

            # Build output structure
            out.append({
                'index': index,
                'id': text_id,
                'polygon': polygon,
            })

        # Sort the surfaces by index
        return sorted(out, key=lambda x: x['index'])


def parse_himanis_volume_name(name):
    """
    Find a volume from a volume name. Specific to Himanis use cases.
    """
    name = name.strip()
    vol_number, vol_suffix = 0, ""
    if not name.isdigit():
        vol_number, vol_suffix = REGEX_VOLUME_NAME.findall(name.upper())[0]
        vol_number = int(vol_number)
    else:
        vol_number = int(name)

    return Element.objects.get(
        Q(type=ElementType.Volume) & (
            Q(name__icontains="JJ{:d}{}".format(vol_number, vol_suffix)) |
            Q(name__icontains="JJ {:d}{}".format(vol_number, vol_suffix)) |
            Q(name__icontains="JJ{:03d}{}".format(vol_number, vol_suffix)) |
            Q(name__icontains="JJ {:03d}{}".format(vol_number, vol_suffix))
        ))


def parse_act_folio(pages, text):
    """
    Get a list of Page instances from folio data in act CSV files and a set of pages to look in
    """
    text = text.strip()
    folios = REGEX_FOLIO.findall(text)
    if len(folios) < 1:
        raise ValueError
    results = []
    for folio in folios:
        # TODO: Handle page suffix
        direction = PageDirection.Verso if folio[2].startswith('v') else PageDirection.Recto
        found_pages = [p for p in pages if p.nb == int(folio[0]) and p.direction == direction]
        if len(found_pages) > 1:
            raise Page.MultipleObjectsReturned
        elif len(found_pages) < 1:
            raise Page.DoesNotExist
        results.append(found_pages[0])
    return results


class SurfaceImporter(object):

    def __init__(self, csvpath=None, xmlpath=None, col_volume=0, col_number=1, col_folio=2, **kwargs):
        assert os.path.isfile(csvpath)
        assert os.path.exists(xmlpath)
        self.xmlpath = xmlpath
        self.csvpath = csvpath

        assert col_volume != col_number != col_folio, 'Column mismatch'

        logger.debug('Loading CSV file')
        CSVRow = namedtuple('CSVRow', 'volume_name, number, folio')
        with open(self.csvpath) as csvfile:
            reader = csv.reader(csvfile, delimiter=';')
            next(reader)  # Skip header
            self.csvdata = [
                CSVRow(
                    row[col_volume],
                    row[col_number],
                    row[col_folio],
                )
                for row in reader
                if row and len(row[0]) > 0
            ]

        self.raw_volume_names = set(row.volume_name for row in self.csvdata)
        self.raw_folios = set((row.volume_name, row.folio) for row in self.csvdata)

    def get_xml_paths(self):
        # Support single file & directories
        if os.path.isdir(self.xmlpath):
            for root, _, filenames in os.walk(self.xmlpath):
                for filename in fnmatch.filter(filenames, "*.xml"):
                    yield os.path.join(root, filename)
        else:
            yield os.path.realpath(self.xmlpath)

    def match_volumes(self):
        logger.debug('Matching volumes with volume names')
        self.volumes = {}
        for volume_name in self.raw_volume_names:
            try:
                self.volumes[volume_name] = parse_himanis_volume_name(volume_name)
                logger.debug("Matched volume '{}' for volume name '{}'".format(self.volumes[volume_name], volume_name))
            except ValueError as e:
                logger.warning('{}\t{}'.format(volume_name, str(e)))
            except Element.DoesNotExist:
                logger.warning('{}\tNo matching volume found'.format(volume_name))
            except Element.MultipleObjectsReturned:
                logger.warning('{}\tMultiple matching volumes found'.format(volume_name))
        if len(self.volumes) < 1:
            raise ValueError("No matched volumes.")

    def match_folios(self):
        logger.debug('Matching folios with pages')
        self.pages = {}
        volume_pages = {}  # Cache for get_descending
        for volume_name, folio in self.raw_folios:
            try:
                volume = self.volumes[volume_name]
            except KeyError:
                logger.warning('Volume not found: {}'.format(volume_name))
                continue
            if volume not in volume_pages:
                volume_pages[volume] = list(
                    Page.objects.get_descending(volume.id)
                                .filter(page_type=PageType.Page)
                                .prefetch_related('zone__image'))
                assert len(volume_pages[volume]) > 0, "{} is empty".format(volume)
            try:
                self.pages[(volume_name, folio)] = parse_act_folio(volume_pages[volume], folio)
                logger.debug("Matched pages {} for folio '{}' in volume {}".format(
                    ", ".join(p.folio for p in self.pages[(volume_name, folio)]),
                    folio, volume_name))
            except ValueError:
                logger.warning("Could not parse folio '{}' in volume '{}'".format(folio, volume_name))
            except Page.DoesNotExist:
                logger.warning("No page found for folio '{}' in volume '{}'".format(folio, volume_name))
            except Page.MultipleObjectsReturned:
                logger.warning("Multiple matching pages found for folio '{}' in volume '{}'".format(folio, volume_name))

        assert self.pages, 'No pages saved'

        # Prepare pages list for XML parsing
        self.pages_list = set(item for sublist in volume_pages.values() for item in sublist)

    def find_page(self, path):
        """
        Get a page whose image corresponds to an image name found in an XML file.
        """
        basename = os.path.splitext(path)[0][:-2]
        return next(
            p
            for p in self.pages_list
            if p.zone and basename in p.zone.image.path
        )

    def parse_xml_files(self, dry=False):
        self.surfaces = {}

        for xml in self.get_xml_paths():
            s = SurfaceParser(xml)
            try:
                page = self.find_page(s.get_metadata()['imageFilename'])
                img = page.zone.image
            except StopIteration:
                logger.warning("No associated image found for XML file '{}'".format(xml))
                continue

            if dry:  # Handle dry run
                logger.info("{}\t{}".format(xml, img.path))
                continue

            if page not in self.surfaces:
                self.surfaces[page] = []

            for surfacedata in s.list_surfaces():
                poly = surfacedata['polygon']
                if poly[0] != poly[-1]:  # Add first point as last point
                    poly.append(poly[0])
                try:
                    z, _ = Zone.objects.get_or_create(image=img, polygon=poly)
                except ValueError as e:
                    logger.warning("Could not import zone with polygon '{}' on image '{}' from file '{}': {}".format(
                        poly, img.id, xml, e))
                except MultipleObjectsReturned:
                    logger.warning('Multiple zones found, picking the first one')
                    z = Zone.objects.filter(image=img, polygon=poly).first()

                surface, _ = Element.objects.get_or_create(
                    type=ElementType.Surface,
                    zone=z,
                    corpus=page.corpus,
                    defaults={'name': "Surface {}".format(surfacedata['id'])})
                self.surfaces[page].append(surface)

        logger.info("Created {} surfaces from XML files".format(sum(len(x) for x in self.surfaces.values())))

    def run(self):
        # Step 1: Find volumes from their names in the CSV file
        self.match_volumes()
        # Step 2: Find pages from folios specified in the CSV file
        self.match_folios()
        # Step 3: Parse XML files and create zones and surface elements
        self.parse_xml_files()

    def dry_run(self):
        self.match_volumes()
        self.match_folios()
        self.parse_xml_files(True)
