from celery import shared_task, group, chain
from celery_once import QueueOnce
from celery.utils.log import get_task_logger
from django.conf import settings
from arkindex.documents.models import Element, Act, Transcription, ElementType, Corpus
from arkindex.documents.indexer import Indexer
from arkindex.documents.importer import URLManifestsImporter, LocalManifestsImporter
from arkindex.documents.surface import SurfaceImporter
from arkindex.documents.surface_link import SurfaceLinker
from arkindex.images.importer import IndexImporter, GallicaIndexImporter
from arkindex.images.models import ImageServer
import os

logger = get_task_logger(__name__)


@shared_task(base=QueueOnce, once={'graceful': True})
def reindex_acts(bulk_size=100, volume_id=None):
    '''
    Reindex all acts
    '''
    indexer = Indexer()
    if volume_id:
        acts = Act.objects.get_descending(volume_id)
    else:
        acts = Act.objects.all()
    indexer.run_index(settings.ES_INDEX_ACTS, Act.INDEX_TYPE, acts, bulk_size=bulk_size)


@shared_task(base=QueueOnce, once={'graceful': True})
def reindex_transcriptions(bulk_size=400, volume_id=None):
    '''
    Reindex all transcriptions
    '''
    indexer = Indexer()
    if volume_id:
        # Lookup all the transcriptions linked to a volume
        transcriptions = Transcription.objects.filter(
            element__in=Element.objects.get_descending(volume_id)
        ).distinct()
    else:
        transcriptions = Transcription.objects.all()
    indexer.run_index(settings.ES_INDEX_TRANSCRIPTIONS, Transcription.INDEX_TYPE, transcriptions, bulk_size=bulk_size)


@shared_task
def import_manifest(path, server_ids=[], corpus_id=None, offline=False, annotations=False, volume_name=None):
    """
    Import a IIIF manifest from a local file or a URL
    """
    try:
        servers = [ImageServer.objects.get(name=server_id) for server_id in server_ids]
    except Exception as e:
        raise ValueError("Image server not found: {}".format(e))

    try:
        corpus = Corpus.objects.get(id=corpus_id)
    except Exception as e:
        raise ValueError("Corpus not found: {}".format(e))

    if any(path.startswith(scheme) for scheme in ('http://', 'https://',)):
        importer_class = URLManifestsImporter
    else:
        importer_class = LocalManifestsImporter

    importer_class(
        servers,
        corpus,
        path,
        offline=offline,
        annotations=annotations,
        volume_name=volume_name,
    ).run()


@shared_task
def import_annotations(path, volume_id, mask=IndexImporter.DEFAULT_MASK):
    """
    Import annotations from .idx.gz files into a volume
    """
    volume = Element.objects.get(id=volume_id, type=ElementType.Volume)
    IndexImporter(path, volume, mask=mask).run()


@shared_task
def import_annotations_csv(name, raw_path, source, index_root):
    """
    Import annotations from a CSV line
    """
    if raw_path.startswith('/home/data/indexes'):
        raw_path = raw_path[18:]
    raw_path = raw_path.lstrip('/')
    index_path = os.path.join(index_root, raw_path)
    assert os.path.exists(index_path), \
        'Index files path not found for volume {} : {}'.format(name, index_path)

    volume = Element.objects.get(name=name, type=ElementType.Volume)

    if source == 'bvmm':
        logger.info("Importing BVMM annotations for volume {}".format(name))
        importer = IndexImporter(index_path, volume, mask=r'(?:.*/)?([^/]+)_[A-Z]\.idx\.gz')
    elif source == 'gallica':
        logger.info("Importing Gallica annotations for volume {}".format(name))
        importer = GallicaIndexImporter(index_path, volume)
    else:  # Try anyway
        logger.info("Importing unknown annotations for volume {}".format(name))
        importer = IndexImporter(index_path, volume)

    importer.run()


@shared_task
def import_manifests_annotations(csv_data, index_root, corpus_id, col_name=0, col_url=1, col_path=2, col_source=3):
    '''
    Import both manifests & annotations
    from a parsed csv file (or whatever table)
    '''
    manifests = [
        import_manifest.si(
            row[col_url], offline=True, annotations=False, volume_name=row[col_name], corpus_id=corpus_id)
        for row in csv_data
    ]
    annotations = [
        import_annotations_csv.si(row[col_name], row[col_path], row[col_source], index_root=index_root)
        for row in csv_data
    ]

    workflow = chain(group(manifests), group(annotations))
    workflow.delay()


@shared_task
def import_surfaces(*args, **kwargs):
    s = SurfaceImporter(*args, **kwargs)
    if kwargs['dry_run']:
        s.dry_run()
    else:
        s.run()


@shared_task
def link_acts_surfaces(volume_id, data, dry_run=False):
    s = SurfaceLinker(volume=Element.objects.get(id=volume_id), data=data)
    if dry_run:
        s.dry_run()
    else:
        s.run()
