from django.urls import reverse
from arkindex.project.tests import FixtureAPITestCase
from rest_framework import status
from arkindex.documents.models import Corpus, Element, ElementType
from django.contrib.auth.models import AnonymousUser


class TestCorpus(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.anon = AnonymousUser()

        cls.corpus_public = Corpus.objects.get(name='Unit Tests', public=True)
        cls.corpus_private = Corpus.objects.create(name='B Private')
        cls.user.corpus_right.create(corpus=cls.corpus_private, user=cls.user, can_write=True)
        cls.corpus_hidden = Corpus.objects.create(name='C Hidden')

    def test_anon(self):
        # An anonymous user has only access to public
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read'],
                    'public': True,
                    'name': 'Unit Tests',
                }
            ]
        )

    def test_user(self):
        # A normal user has access to public + its private
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    'id': str(self.corpus_private.id),
                    'rights': ['read', 'write'],
                    'public': False,
                    'name': 'B Private',
                },
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': True,
                    'name': 'Unit Tests',
                }
            ]
        )

    def test_admin(self):
        # An admin has access to everything
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    'id': str(c.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': c.public,
                    'name': c.name,
                }
                for c in Corpus.objects.all().order_by('name')
            ]
        )

    def test_mixin(self):
        vol1 = Element.objects.get(name="Volume 1", type=ElementType.Volume)
        vol2 = Element.objects.get(name="Volume 2", type=ElementType.Volume)

        response = self.client.get('{}?corpus={}'.format(reverse('api:elements'), str(self.corpus_private.id)))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        response = self.client.get('{}?corpus={}'.format(reverse('api:elements'), str(self.corpus_public.id)))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 2)
        self.assertCountEqual(
            [r['id'] for r in data['results']],
            [str(vol1.id), str(vol2.id)],
        )
