from unittest.mock import patch
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.surface_link import ParsedFolio, SurfaceLinker
from arkindex.documents.models import Element, ElementType


class TestSurfaceLinker(FixtureTestCase):
    """Test surface-act linking process."""

    def setUp(self):
        self.vol = Element.objects.get(name="Volume 1")
        self.sa = Element.objects.get(name="Surface A")
        self.sb = Element.objects.get(name="Surface B")
        self.sc = Element.objects.get(name="Surface C")
        self.sd = Element.objects.get(name="Surface D")
        self.se = Element.objects.get(name="Surface E")
        self.sf = Element.objects.get(name="Surface F")

    def test_init(self):
        "Test SurfaceLinker performs checks on its inputs"

        with self.assertRaises(Exception):
            SurfaceLinker({"this is": not "a volume"}, [('1', '1r'), ('2', '1v')])

        with self.assertRaises(Exception):
            SurfaceLinker(
                Element(type=ElementType.Line, name="not a volume"),
                [('1', '1r'), ('2', '1v')])

        with self.assertRaises(Exception):
            SurfaceLinker(self.vol, "bad data")

        with self.assertRaises(Exception):
            SurfaceLinker(self.vol, ['still', 'bad'])

        with self.assertRaises(Exception):
            SurfaceLinker(self.vol, [('still', 'not'), ['good']])

        with self.assertRaises(Exception):
            SurfaceLinker(self.vol, [('almost', 'there', 'but'), ['still', 'needs', 'efforts']])

    def test_pop_surfaces(self):
        "Test SurfaceLinker's _pop_surfaces on valid folios"

        s = SurfaceLinker(self.vol, [])
        s.find_pages()
        self.assertEqual(len(s.pages), 3)
        s.find_surfaces()
        self.assertEqual(s.surfaces_count, 6)

        folio = ParsedFolio.from_raw('1', None, 'r')
        surfaces = s._pop_surfaces(folio)
        self.assertListEqual(surfaces, [self.sa, self.sb])
        self.assertEqual(len(s.surfaces[folio]), 0)

        folio = ParsedFolio.from_raw('2', None, 'r')
        surfaces = s._pop_surfaces(folio, index=0)
        self.assertListEqual(surfaces, [self.sd])
        self.assertEqual(len(s.surfaces[folio]), 2)

    @patch('arkindex.documents.surface_link.logger')
    def test_pop_surfaces_warning(self, mock_logger):
        "Test SurfaceLinker's _pop_surfaces warnings with wrong folios or lack of surfaces"

        s = SurfaceLinker(self.vol, [])
        s.find_pages()
        self.assertEqual(len(s.pages), 3)
        s.find_surfaces()
        self.assertEqual(s.surfaces_count, 6)
        mock_logger.reset_mock()

        surfaces = s._pop_surfaces(ParsedFolio.from_raw('5', None, ''))
        self.assertEqual(surfaces, [])
        self.assertEqual(mock_logger.warning.call_count, 1)
        mock_logger.reset_mock()

        folio = ParsedFolio.from_raw('1', None, 'v')
        s._pop_surfaces(folio)

        surfaces = s._pop_surfaces(folio)
        self.assertEqual(surfaces, [])
        self.assertEqual(mock_logger.warning.call_count, 1)
        mock_logger.reset_mock()

        surfaces = s._pop_surfaces(folio, index=4)
        self.assertEqual(surfaces, [])
        self.assertEqual(mock_logger.warning.call_count, 1)
        mock_logger.reset_mock()

    def test_one_folio(self):
        """
        Test surface matching for an act with 2 folios on same page
        Greedy test case
        """
        self.assertTrue(self.sa.zone.polygon.center.y < self.sb.zone.polygon.center.y)
        s = SurfaceLinker(self.vol, [('1', '001r'), ('2', '001r')])
        s.find_pages()
        self.assertEqual(len(s.pages), 3)
        s.find_surfaces()
        self.assertEqual(s.surfaces_count, 6)
        s.find_acts()
        s.link_acts_folios()
        s.resolve_conflicts()
        s.link_surfaces()
        self.assertEqual(s.parsed_acts, 2)
        self.assertEqual(s.linked_surfaces, 2)
        self.assertIn('1', s.acts_surfaces)
        self.assertListEqual(s.acts_surfaces['1'], [self.sa])
        self.assertIn('2', s.acts_surfaces)
        self.assertListEqual(s.acts_surfaces['2'], [self.sb])

    def test_two_folios(self):
        """
        Test surface matching for an act with two folios
        End of page 1 + start of page 2
        """
        s = SurfaceLinker(self.vol, [('2', '1r-1v')])
        s.find_pages()
        self.assertEqual(len(s.pages), 3)
        s.find_surfaces()
        self.assertEqual(s.surfaces_count, 6)
        s.find_acts()
        s.link_acts_folios()
        s.resolve_conflicts()
        s.link_surfaces()
        self.assertEqual(s.parsed_acts, 1)
        self.assertEqual(s.linked_surfaces, 2)
        self.assertIn('2', s.acts_surfaces)
        self.assertListEqual(s.acts_surfaces['2'], [self.sb, self.sc])

    def test_three_folios(self):
        """
        Test surface matching for an act with three folios
        End of page 1 + full pages 2 + start of page 3
        """
        s = SurfaceLinker(self.vol, [('3', '1r-2r')])
        s.find_pages()
        self.assertEqual(len(s.pages), 3)
        s.find_surfaces()
        self.assertEqual(s.surfaces_count, 6)
        s.find_surfaces()
        s.find_acts()
        s.link_acts_folios()
        s.resolve_conflicts()
        s.link_surfaces()
        self.assertEqual(s.parsed_acts, 1)
        self.assertEqual(s.linked_surfaces, 3)
        self.assertIn('3', s.acts_surfaces)
        self.assertListEqual(s.acts_surfaces['3'], [self.sb, self.sc, self.sd])

    def test_wrong_folio(self):
        """
        Test surface matching for an act with an invalid folio
        No pages
        """
        s = SurfaceLinker(self.vol, [('4', '5v')])
        s.find_pages()
        s.find_surfaces()
        s.find_acts()
        s.link_acts_folios()
        s.resolve_conflicts()
        s.link_surfaces()
        self.assertEqual(s.parsed_acts, 1)
        self.assertEqual(s.linked_surfaces, 0)
        self.assertIn('4', s.acts_surfaces)
        self.assertEqual(s.acts_surfaces['4'], [])

    def test_find_missing_acts(self):
        """
        Test automatic detection of missing acts
        """
        s = SurfaceLinker(self.vol, [('2', '1v'), ('4', '2r')])
        s.find_missing_acts()
        self.assertListEqual(s.acts_raw, [('1', ''), ('2', '1v'), ('3', ''), ('4', '2r')])

    def test_conflict(self):
        """
        Test surface matching with a conflict
        """
        s = SurfaceLinker(self.vol, [('1', '1r'), ('2', '1r-1v'), ('3', '2r'), ('4', '2r'), ('5', '2r')])
        s.find_pages()
        s.find_surfaces()
        s.find_acts()
        s.link_acts_folios()
        s.resolve_conflicts()
        s.link_surfaces()
        self.assertEqual(s.parsed_acts, 5)
        self.assertEqual(s.linked_surfaces, 6)
        self.assertDictEqual(s.acts_surfaces, {
            '1': [self.sa],
            '2': [self.sb, self.sc],
            '3': [self.sd],
            '4': [self.se],
            '5': [self.sf],
        })
