from lxml import etree
from arkindex.documents.models import Act
from arkindex.documents.tei import Text, TeiParser
from arkindex.dataimport.models import Repository, Revision, DataImportFailure
from arkindex.users.models import OAuthCredentials
from arkindex.project.tests import FixtureTestCase
import os.path

FIXTURES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'tei_samples',
)


class TestTeiElement(FixtureTestCase):

    def setUp(self):
        self.repo = Repository.objects.create(
            url='http://repo',
            hook_token='token',
            corpus=self.corpus,
            credentials=OAuthCredentials.objects.create(
                user=self.user,
                provider_name='provider',
                provider_url='https://somewhere',
            ),
        )

    def test_apply_xslt(self):
        tree = etree.parse(os.path.join(FIXTURES, 'arguments.xml'))
        te = Text(tree.getroot())
        self.assertMultiLineEqual(
            # Remove indentations
            "".join(line.strip() for line in te.argument.splitlines()),
            '<div xmlns:tei="http://www.tei-c.org/ns/1.0" class="argument">'
            '<p>Argument</p>'
            '<p><em>Italique</em></p>'
            '<p><sup>Supérieur</sup></p>'
            '<p><sup>Supérieur</sup></p>'
            '<p><blockquote>Citation</blockquote></p>'
            '</div>')

    def test_save_update(self):
        """
        Check saving metadata for a TeiElement updates existing data
        """
        act = Act.objects.get(number="1")
        tree_before = etree.parse(os.path.join(FIXTURES, 'update_before.xml'))
        tree_after = etree.parse(os.path.join(FIXTURES, 'update_after.xml'))
        te_before = Text(tree_before.getroot())
        te_after = Text(tree_after.getroot())

        rev1 = Revision.objects.create(repo=self.repo, hash='42', ref='ref/heads/master', message='a', author='me')
        rev2 = Revision.objects.create(repo=self.repo, hash='43', ref='ref/heads/master', message='b', author='me')

        te_before.save(act, rev1)
        location = act.metadatas.get(name="location")
        persons = act.metadatas.get(name="persons")
        self.assertEqual(location.value, 'Mars')
        self.assertEqual(location.revision, rev1)
        self.assertEqual(persons.value, 'Someone')
        self.assertEqual(act.metadatas.get(name="places").value, 'Somewhere')
        self.assertFalse(act.metadatas.filter(name="subjects").exists())

        # Create subject, update person and delete place, leave location untouched
        te_after.save(act, rev2)
        location.refresh_from_db()
        persons.refresh_from_db()
        self.assertEqual(location.value, 'Mars')
        self.assertEqual(location.revision, rev1)
        self.assertEqual(persons.value, 'Someone else')
        self.assertEqual(persons.revision, rev2)
        self.assertFalse(act.metadatas.filter(name="places").exists())
        self.assertEqual(act.metadatas.get(name="subjects").value, 'Something')

    def test_match_error(self):
        xml_path = os.path.join(FIXTURES, 'arguments.xml')
        parser = TeiParser(xml_path)
        parser._match_error(parser.corpus, 'Some funny test message')
        self.assertEqual(len(parser.match_errors), 1)
        failure = parser.match_errors[0]
        self.assertIsInstance(failure, DataImportFailure)
        self.assertEqual(failure.path, xml_path)
        self.assertEqual(failure.line, 1)
        self.assertEqual(failure.message, 'Some funny test message')
        self.assertIsNone(failure.dataimport_id)
        self.assertIsNone(failure.element_id)
