from django.urls import reverse
from unittest.mock import patch
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.polygon import Polygon
from arkindex.documents.models import Page, Transcription, TranscriptionType
import uuid


class TestTranscriptionCreate(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    def setUp(self):
        self.page = Page.objects.get(corpus=self.corpus, zone__image__path='img1')

    def test_require_login(self):
        response = self.client.post(reverse('api:transcription-create'), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.documents.api.transcriptions.Indexer')
    def test_create_transcription(self, indexer):
        """
        Checks the view creates transcriptions, zones, links, paths and runs ES indexing
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:transcription-create'), format='json', data={
            "type": "word",
            "element": str(self.page.id),
            "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
            "text": "NEKUDOTAYIM",
            "score": 0.83,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_ts = Transcription.objects.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(new_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(new_ts.score, 0.83)
        self.assertTrue(self.page.transcriptions.filter(pk=new_ts.id).exists())
        self.assertTrue(indexer.return_value.run_index.called)

    def test_unique_zone(self):
        """
        Checks the view reuses zones when available
        """
        self.client.force_login(self.user)
        ts = Transcription.objects.get(zone__image__path='img1', text="PARIS")
        response = self.client.post(reverse('api:transcription-create'), format='json', data={
            "type": "word",
            "element": str(self.page.id),
            "polygon": ts.zone.polygon.serialize(),
            "text": "GLOUBIBOULGA",
            "score": 0.8,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            Transcription.objects.get(text="GLOUBIBOULGA").zone.id, ts.zone.id)

    def test_update_transcription(self):
        """
        Checks the view updates transcriptions when they already exist
        """
        self.client.force_login(self.user)
        ts = Transcription.objects.get(zone__image__path='img1', text="PARIS")
        self.assertNotEqual(ts.score, 0.99)
        response = self.client.post(reverse('api:transcription-create'), format='json', data={
            "type": "word",
            "element": str(self.page.id),
            "polygon": ts.zone.polygon.serialize(),
            "text": ts.text,
            "score": 0.99,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Transcription.objects.filter(zone__image__path='img1', text=ts.text).count(), 1)
        ts.refresh_from_db()
        self.assertEqual(ts.score, 0.99)

    def test_invalid_data(self):
        """
        Checks the view validates data properly
        (score between 0 and 1, minimum points in polygon, non-negative coordinates, non-existent element)
        """
        self.client.force_login(self.user)

        post_data = {
            "type": "word",
            "element": str(self.page.id),
            "polygon": [(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)],
            "text": "NEKUDOTAYIM",
            "score": 0.83,
        }

        # Negative score
        post_data['score'] = -1
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Score over 100%
        post_data['score'] = 2
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Missing element
        post_data['score'] = 0.83
        post_data['element'] = str(uuid.uuid4())
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Not enough polygon points
        post_data['polygon'] = [(0, 0), (100, 100)]
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex.documents.api.transcriptions.Indexer')
    def test_create_bulk_transcription(self, indexer):
        """
        Checks the view creates transcriptions, zones, links, paths and runs ES indexing
        Using bulk_transcriptions
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data={
            "parent": str(self.page.id),
            "image": str(self.page.zone.image.id),
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_ts = Transcription.objects.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(new_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(new_ts.score, 0.83)
        self.assertTrue(self.page.transcriptions.filter(pk=new_ts.id).exists())

        # indexer called
        self.assertTrue(indexer.return_value.run_index.called)
