from django.urls import reverse
from arkindex.project.tests import FixtureAPITestCase
from rest_framework import status
from tripoli import IIIFValidator
from arkindex.documents.models import Element, ElementType


class TestVolumeManifestSerializer(FixtureAPITestCase):
    """Tests for VolumeManifestSerializer class"""

    def test_normal_manifest(self):
        # Just a normal situation.
        vol = Element.objects.get(type=ElementType.Volume, name="Volume 1")
        response = self.client.get(reverse('api:volume-manifest', kwargs={'pk': vol.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        manifest = response.json()

        self.assertIn("@context", manifest)
        self.assertIn("@id", manifest)
        self.assertIn("@type", manifest)
        self.assertIn("thumbnail", manifest)
        self.assertIn("structures", manifest)
        self.assertIn("sequences", manifest)
        self.assertIn("label", manifest)

        self.assertEqual(manifest['@type'], "sc:Manifest")
        self.assertEqual(manifest['@context'],
                         "http://iiif.io/api/presentation/2/context.json")
        self.assertEqual(manifest['label'], vol.name)
        self.assertEqual(manifest['thumbnail']['service']['@context'],
                         "http://iiif.io/api/image/2/context.json")
        self.assertEqual(manifest['thumbnail']['service']['profile'],
                         "http://iiif.io/api/image/2/level2.json")
        self.assertIn('server/img', manifest['thumbnail']['service']['@id'])

        self.assertEqual(len(manifest['structures']), 4)
        topstruct = next(s for s in manifest['structures'] if s['viewingHint'] == "top")
        self.assertIsNotNone(topstruct)
        self.assertEqual(topstruct['label'], vol.name)
        self.assertEqual(topstruct['@type'], 'sc:Range')
        self.assertEqual(len(topstruct['ranges']), 3)
        self.assertEqual(manifest['@id'], topstruct['@id'])

        self.assertEqual(len(manifest['sequences']), 1)
        self.assertEqual(manifest['sequences'][0]['@type'], 'sc:Sequence')
        canvases = manifest['sequences'][0]['canvases']
        self.assertEqual(len(canvases), 3)
        for canvas in canvases:
            self.assertIn('@type', canvas)
            self.assertIn('@id', canvas)
            self.assertIn('label', canvas)
            self.assertIn('width', canvas)
            self.assertIn('height', canvas)
            self.assertIn('images', canvas)
            self.assertEqual(canvas['@type'], 'sc:Canvas')
            self.assertEqual(len(canvas['images']), 1)

        images = [c['images'][0] for c in canvases]
        for image in images:
            self.assertIn('@type', image)
            self.assertIn('resource', image)
            self.assertIn('on', image)
            self.assertIn('motivation', image)
            self.assertEqual(image['@type'], 'oa:Annotation')
            self.assertEqual(image['motivation'], 'sc:painting')
            res = image['resource']
            self.assertIn('width', res)
            self.assertIn('height', res)
            self.assertIn('service', res)
            self.assertEqual(res['@type'], 'dctypes:Image')
            self.assertEqual(res['service']['@context'],
                             "http://iiif.io/api/image/2/context.json")
            self.assertEqual(res['service']['profile'],
                             "http://iiif.io/api/image/2/level2.json")

    def test_no_page(self):
        # A manifest for an empty volume
        response = self.client.get(reverse('api:volume-manifest', kwargs={
            'pk': Element.objects.create(corpus=self.corpus, type=ElementType.Volume, name="Empty Volume").id
        }))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        manifest = response.json()

        self.assertEqual(manifest['thumbnail'], None)
        self.assertEqual(len(manifest['structures']), 1)
        topstruct = manifest['structures'][0]
        self.assertEqual(len(topstruct['ranges']), 0)
        self.assertEqual(len(manifest['sequences']), 1)
        self.assertEqual(len(manifest['sequences'][0]['canvases']), 0)

    def test_normal_manifest_iiif_validation(self):
        vol = Element.objects.get(type=ElementType.Volume, name="Volume 1")
        response = self.client.get(reverse('api:volume-manifest', kwargs={'pk': vol.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        manifest = response.json()
        iv = IIIFValidator()
        iv.validate(manifest)
        self.assertTrue(iv.is_valid)
