from arkindex.images.models import Zone, Image
from arkindex.documents.models import Transcription, TranscriptionType, Element, Page, ElementType
from arkindex.project.polygon import Polygon
from collections import namedtuple
from abc import ABC, abstractmethod
from django.db import transaction
import os
import re
import gzip
import logging
import fnmatch
import uuid

REGEX_INDEX = re.compile(
    b'^(?:line_(\d+) )?(.+) \d+ ([\de\-\.]+) (\d+) (\d+) (\d+) (\d+)')

logger = logging.getLogger(__name__)


def import_indexes(image, page, index_path, extension='jpg'):
    """
    Import indexes from file
    One gzipped index file per image
    Format: #line word n/a score-[0,1] x1 y1 x2 y2
    """
    assert isinstance(image, Image)
    assert isinstance(page, Page)
    assert os.path.exists(index_path), \
        'Missing index {}'.format(index_path)
    assert index_path.endswith('.idx.gz')

    # Parse lines content
    lines = []
    with gzip.open(index_path) as f:
        for line in f.readlines():
            if line.decode('utf-8').startswith('#'):
                continue
            index = REGEX_INDEX.match(line)
            if index is None:
                logger.warning('Index parsing failed : {}'.format(line))
                continue

            # Build zone
            data = index.groups()
            x, y = int(data[3]), int(data[4])
            lines.append({
                'x': x,
                'y': y,
                'width': int(data[5]) - x,
                'height': int(data[6]) - y,
                'text': data[1].decode('utf-8'),
                'line': int(data[0] or 0),
                'score': float(data[2]),
            })

    logger.info('Parsed {} lines'.format(len(lines)))

    # Create transcriptions
    new_transcriptions = bulk_transcriptions(image, page, lines)
    logger.info('Created {} transcriptions '.format(len(new_transcriptions)))


def bulk_transcriptions(image, parent, items):
    """
    Create transcriptions and zones in bulk
    """
    assert isinstance(parent, Element)

    # Link a transcription data with a Polygon
    # This is hashable (box is hashable)
    TrPolygon = namedtuple('TrPolygon', 'type, polygon, text, score')

    # Build all TrPolygon from items
    required = {
        TrPolygon(
            i.get('type', TranscriptionType.Word),
            Polygon.from_dict(i),
            i['text'],
            float(i['score']),
        )
        for i in items
    }

    # List all zones in image
    zones = {
        z.id: z.polygon
        for z in image.zones.all()
    }

    # Build all TrPolygon from existing
    existing = {
        TrPolygon(
            tr.type,
            tr.zone.polygon,
            tr.text,
            tr.score,
        )
        for tr in parent.transcriptions.all().prefetch_related('zone')
    }

    # Calc needed TrPolygon to build
    needed = required.difference(existing)
    if not needed:
        return []

    def new_transcription(tr_polygon):
        zone = Zone(
            id=uuid.uuid4(),
            image=image,
            polygon=tr_polygon.polygon.serialize(),
        )
        tr = Transcription(
            element_id=parent.id,
            zone_id=zone.id,
            type=tr_polygon.type,
            text=tr_polygon.text,
            score=tr_polygon.score,
        )
        return (tr, zone)

    with transaction.atomic():

        # Build transcriptions & zones instances at the same time
        transcriptions, zones = zip(*map(new_transcription, needed))

        # Create zones in bulk
        Zone.objects.bulk_create(zones)

        # Create transcriptions in bulk
        # with assigned ids
        Transcription.objects.bulk_create(transcriptions)

    return transcriptions


class BaseIndexImporter(ABC):
    """Import index files (.idx.gz) as transcriptions."""

    def __init__(self, path, volume):
        assert os.path.exists(path)
        assert isinstance(volume, Element)
        assert volume.type == ElementType.Volume
        self.path = path
        self.volume = volume

        logger.debug('Fetching pages for volume {}'.format(str(self.volume)))
        self.pages = list(Page.objects.get_descending(self.volume.id).prefetch_related('zone', 'zone__image'))

    def get_index_paths(self):
        # Support single file & directories
        if os.path.isdir(self.path):
            for root, _, filenames in os.walk(self.path):
                for filename in fnmatch.filter(filenames, "*.idx.gz"):
                    yield os.path.join(root, filename)
        else:
            yield os.path.realpath(self.path)

    @abstractmethod
    def get_image(self, path):
        "Return an Image instance for a given index file path"

    def get_page(self, image):
        assert isinstance(image, Image)
        try:
            return next(p for p in self.pages if p.zone.image == image)
        except StopIteration:
            raise Page.DoesNotExist

    def run(self):
        for index_path in self.get_index_paths():
            logger.info("Parsing index file {}".format(index_path))
            try:
                image = self.get_image(index_path)
                page = self.get_page(image)
                import_indexes(image, page, index_path)
            except Image.DoesNotExist:
                logger.warning("No associated image found for file {}".format(index_path))
            except Page.DoesNotExist:
                logger.warning("No associated page found for file {}".format(index_path))
            except Image.MultipleObjectsReturned:
                logger.warning("Multiple associated images found for file {}".format(index_path))

    def dry_run(self):
        for index_path in self.get_index_paths():
            image = None
            try:
                image = self.get_image(index_path)
            except (Image.DoesNotExist, Image.MultipleObjectsReturned):
                pass
            if image is None:
                logger.warning("{}\tFAIL".format(index_path))
            else:
                logger.info("{}\t{}".format(index_path, image.path))


class IndexImporter(BaseIndexImporter):
    """Basic index importer with image matching based on a regular expression."""

    DEFAULT_MASK = r'(?:.*/)?([^/]+)(?:_[a-zA-Z]*)?\.idx\.gz'

    def __init__(self, path, volume, mask=DEFAULT_MASK):
        self.mask = re.compile(mask)
        logger.debug('Using mask {}'.format(self.mask.pattern))
        super().__init__(path, volume)
        self.images = [p.zone.image for p in self.pages]

    def get_image(self, path):
        try:
            image_id = self.mask.findall(path)[0]
            logger.debug('Matched {} for path {}'.format(image_id, path))
        except IndexError:
            logger.debug('Mask did not match path {}'.format(path))
            raise Image.DoesNotExist
        try:
            return next(img for img in self.images if image_id in img.path)
        except StopIteration:
            raise Image.DoesNotExist


class GallicaIndexImporter(BaseIndexImporter):
    """Special importer due to Gallica's complicated URLs"""

    REGEX = re.compile(r'.*_0*([0-9]+)(?:_[a-z]+)?\.idx\.gz')

    def __init__(self, path, volume):
        super().__init__(path, volume)
        self.images = {
            p.zone.image.path.rpartition('/')[2]: p.zone.image
            for p in self.pages
        }

    def get_image(self, path):
        try:
            return self.images['f' + GallicaIndexImporter.REGEX.findall(path)[0]]
        except IndexError:
            logger.debug('Mask did not match path {}'.format(path))
            raise Image.DoesNotExist
        except KeyError:
            raise Image.DoesNotExist
