from django.test import TestCase
from fakeredis import FakeStrictRedis
from rest_framework.test import APITestCase
from unittest.mock import patch
from arkindex.documents.models import Corpus
from arkindex.images.models import ImageServer
from arkindex.users.models import User


class FixtureMixin(object):
    """
    Add the database fixture to a test case
    """

    fixtures = ['data.json', ]

    @classmethod
    def setUpTestData(cls):
        cls.corpus = Corpus.objects.get(name='Unit Tests')
        cls.user = User.objects.get(email='user@user.fr')
        cls.superuser = User.objects.get(email='root@root.fr')
        cls.imgsrv = ImageServer.objects.get(url='http://server')


class FixtureTestCase(FixtureMixin, TestCase):
    """
    Django test case with the database fixture
    """


class FixtureAPITestCase(FixtureMixin, APITestCase):
    """
    Django REST Framework test case with the database fixture
    """


class RedisMockMixin(object):
    """
    Add Redis mocking to a test case
    """

    def setUp(self):
        self.patches = [patch(x) for x in [
            'celery.backends.redis.RedisBackend._create_client',
            'kombu.transport.redis.Channel._create_client',
            'celery_once.backends.redis.get_redis',
        ]]
        self.mocked = [p.start() for p in self.patches]
        self.redis = FakeStrictRedis()
        for m in self.mocked:
            m.return_value = self.redis

        # Patch the add_message
        self.messages = patch('arkindex.project.celery.ExtendedRedisBackend.add_message')
        self.messages.start()

    def tearDown(self):
        for p in self.patches:
            p.stop()
        self.messages.stop()
        self.redis.flushall()


class RedisMockTestCase(RedisMockMixin, TestCase):
    """
    Django test case with Redis mocking
    """


class RedisMockAPITestCase(RedisMockMixin, APITestCase):
    """
    Django REST Framework test case with Redis mocking
    """
