from django.urls import reverse
from django.core.mail import send_mail
from django.contrib.auth import login, logout
from django.contrib.auth.tokens import default_token_generator
from django.contrib.auth.models import Group
from django.views.generic import RedirectView
from django.template.loader import render_to_string
from django.utils.http import urlsafe_base64_encode
from rest_framework.generics import ListAPIView, RetrieveDestroyAPIView, RetrieveUpdateDestroyAPIView, CreateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.exceptions import AuthenticationFailed, ValidationError
from arkindex.documents.models import Corpus, Element, ElementType
from arkindex.users.providers import oauth_providers
from arkindex.users.models import User
from arkindex.users.serializers import \
    OAuthCredentialsSerializer, OAuthProviderClassSerializer, UserSerializer, NewUserSerializer, EmailLoginSerializer
import urllib.parse
import logging

logger = logging.getLogger(__name__)


class ProvidersList(ListAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = OAuthProviderClassSerializer
    pagination_class = None

    def get_queryset(self):
        return list(filter(lambda p: p.enabled(), oauth_providers))


class CredentialsList(ListAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = OAuthCredentialsSerializer

    def get_queryset(self):
        return self.request.user.credentials.exclude(token=None).order_by('id')


class CredentialsRetrieve(RetrieveDestroyAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = OAuthCredentialsSerializer

    def get_queryset(self):
        return self.request.user.credentials.exclude(token=None).order_by('id')

    def perform_destroy(self, instance):
        instance.provider_class(request=self.request, credentials=instance).disconnect()
        super().perform_destroy(instance)


class UserRetrieve(RetrieveUpdateDestroyAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = UserSerializer

    def get_object(self):
        return self.request.user

    def perform_destroy(self, instance):
        logout(self.request)


class UserCreate(CreateAPIView):
    serializer_class = NewUserSerializer

    def perform_create(self, serializer):
        user = serializer.save()
        demo_group, _ = Group.objects.get_or_create(name="Demo")
        user.groups.add(demo_group)
        login(self.request, user)

        corpus = Corpus.objects.create(name=user.email)
        corpus.corpus_right.create(user=user, can_write=True, can_admin=True)
        Element.objects.create(type=ElementType.Volume, name='Demo', corpus=corpus)

        activation_url = '{}?{}'.format(
            self.request.build_absolute_uri(reverse('api:user-token')),
            urllib.parse.urlencode({
                'email': user.email,
                'token': default_token_generator.make_token(user),
            }),
        )
        sent = send_mail(
            # Subject cannot have any newlines
            subject=''.join(render_to_string('registration/verification_email_subject.txt').splitlines()),
            message=render_to_string(
                'registration/verification_email.html',
                context={
                    'url': activation_url,
                },
                request=self.request,
            ),
            from_email=None,
            recipient_list=[user.email],
            fail_silently=True,
        )
        if sent == 0:
            logger.error('Failed to send registration email to {}'.format(user.email))


class UserEmailLogin(CreateAPIView):
    serializer_class = EmailLoginSerializer

    def perform_create(self, serializer):
        try:
            user = User.objects.get_by_natural_key(serializer.validated_data['email'])
            assert user.has_usable_password()
            assert user.check_password(serializer.validated_data['password'])
        except (User.DoesNotExist, AssertionError):
            raise AuthenticationFailed()
        login(self.request, user)


class UserEmailVerification(RedirectView):

    def get_redirect_url(self):
        if not all(arg in self.request.GET for arg in ('email', 'token')):
            raise ValidationError()
        try:
            user = User.objects.get_by_natural_key(self.request.GET['email'])
            assert default_token_generator.check_token(user, self.request.GET['token'])
        except (User.DoesNotExist, AssertionError):
            raise AuthenticationFailed()

        user.verified_email = True
        user.save()

        if user.has_usable_password():
            login(self.request, user)
            return '/'

        return reverse('password_reset_confirm', kwargs={
            'uidb64': urlsafe_base64_encode(str(user.id).encode()).decode(),
            'token': self.request.GET['token'],
        })
